# -*- coding: utf-8 -*-
"""
Created on Sat Jan  8 16:40:18 2022

@author: richa
"""
import datetime
import logging
import os
import sys


class Logger():
    """Logger context manager"""

    def __init__(self, folder, filename):
        self.folder = folder
        self.filename = filename
        self.logger = None

    def __enter__(self):
        self._init_folder()
        self._init_logger()
        return self.logger

    def __exit__(self, *_):
        self.logger.handlers.clear()
        logging.shutdown()

    def _init_logger(self):
        logging.basicConfig(format='%(message)s', level=logging.DEBUG)
        self.logger = logging.getLogger('game.log')
        self.logger.handlers.clear()

        now = datetime.datetime.now().strftime('%y%m%d_%H%M%S')
        filepath = os.path.join(self.folder, f'{now}_game.log')
        file_handler = logging.FileHandler(filepath, mode='w')
        file_handler.setLevel(logging.DEBUG)
        self.logger.addHandler(file_handler)

        self.logger.addHandler(logging.StreamHandler(sys.stdout))
        self.logger.propagate = False

    def _init_folder(self):
        if not os.path.isdir(self.folder):
            os.makedirs(self.folder)



# -*- coding: utf-8 -*-
"""
Created on Fri Jan  7 14:33:33 2022

@author: richa
"""
import math
from dataclasses import dataclass
from typing import Callable
from typing import Dict
from typing import List
from typing import Tuple

import numpy as np

from .board import Board
from .hashing import compute_hash
from .piece import Piece

@dataclass
class ReversibleMove:
    """Context manager move that reverses board state to initial state on exit"""

    board: Board
    piece: Piece
    position: Tuple[int, int]
    enemy_pieces: List[Piece]

    def __post_init__(self):
        self.original_position = None
        self.captured_piece = None

    def __enter__(self):
        self.original_position = self.piece.position
        self.captured_piece = self.board.move_piece_and_capture(
            self.position, self.piece, self.enemy_pieces, log=False
        )
        return self

    def __exit__(self, *_):
        # undo move
        self.board.move_piece_and_capture(
            self.original_position, self.piece, self.enemy_pieces, log=False
        )
        self.piece.position_history.pop()
        self.piece.position_history.pop()
        self.piece.update(self.board)
        self.board[self.position] = self.captured_piece
        self.board.move_history.pop()
        self.board.move_history.pop()
        if self.captured_piece is not None:
            self.captured_piece.captured = False
            self.enemy_pieces.append(self.captured_piece)


def evaluate_length(board, team, enemy):
    """Evaluates the board state based on the amount of moves allied pieces
    and enemy pieces can make.
    """
    #HACK: using compute_all_moves instead of compute_valid_moves to save computation time
    return len(team.compute_all_moves(board)) - len(enemy.compute_all_moves(board))

def evaluate_length_with_captures(board, team, enemy):
    """Evaluates the board state based on the amount of moves allied pieces
    and enemy pieces can make.
    """
    #HACK: using compute_all_moves instead of compute_valid_moves to save computation time
    ally_moves = sum([2 if move.can_capture else 1 for _, move in team.compute_all_moves(board)])
    enemy_moves = sum([3 if move.can_capture else 1 for _, move in enemy.compute_all_moves(board)])
    return (ally_moves - 1) / enemy_moves if enemy_moves else math.inf

def mating_strategy(board, team, enemy):
    """Incentivizes disabling the enemy kings movements while keeping own movement high"""
    return (len(team.compute_valid_moves(board, enemy.pieces))
            - len(enemy.king.compute_valid_moves(board)))

def evaluate_distance(board, team, enemy):
    """Evaluates board state based on the closeness to the enemy king"""
    #HACK: using compute_all_moves instead of compute_valid_moves to save computation time
    #pylint: disable=invalid-name
    def inverse_distance(point1, point2):
        x1, y1 = point1
        x2, y2 = point2
        value = math.sqrt((x1 - x2) ** 2 + abs(y1 - y2) ** 2)
        if value == 0:
            return 1
        return 1 / value
    sum_ally = sum(inverse_distance(enemy.king.position, move.position)
                   for _, move in team.compute_all_moves(board))
    sum_enemy = sum(inverse_distance(team.king.position, move.position)
                    for _, move in enemy.compute_all_moves(board))
    return sum_ally - sum_enemy

def evaluate_distance_np(board, team, enemy):
    """Evaluates board state based on the closeness to the enemy king"""
    #seems to lead to very wonky games...
    ally_moves = np.array([move.position for _, move in team.compute_all_moves(board)])
    enemy_moves = np.array([move.position for _, move in enemy.compute_all_moves(board)])
    king_position = np.array([team.king.position])
    enemy_king_position = np.array([enemy.king.position])
    ally_distances = np.linalg.norm(enemy_king_position - ally_moves)
    enemy_distances = np.linalg.norm(king_position - enemy_moves)
    return enemy_distances - ally_distances

@dataclass
class Minimax:
    """Minimax algorithm class with alpha-beta pruning and customizable evaluation"""

    evaluation_function: Callable
    hash_values: Dict

    def __post_init__(self):
        self.board_hashes = {}

    #pylint: disable=too-many-arguments,too-many-locals,too-many-branches #for now...
    def run(self, board, team, enemy, depth, alpha, beta, maximizing_player):
        """Minimax algorithm with alpha-beta pruning.
        At depth=3, computation speed is still relatively fast.
        At depth=4, it slows down considerably, but does make much better moves.
        """

        if board.is_draw_by_repetition():
            return 0, None

        if depth == 0: #or game over
            evaluation = self.evaluation_function(board, team, enemy)
            hash_ = compute_hash(board, self.hash_values)
            self.board_hashes[maximizing_player, hash_] = evaluation
            return evaluation, None

        if maximizing_player:
            best_move = None
            max_eval = -math.inf
            for piece, move in team.compute_valid_moves(board, enemy.pieces):
                search_depth = depth - 1
                search_depth = piece.increase_search_depth(search_depth)

                hash_ = compute_hash(board, self.hash_values)
                eval_position = self.board_hashes.get((maximizing_player, hash_))
                if eval_position is None:
                    with ReversibleMove(board, piece, move.position, enemy.pieces):
                        eval_position = self.run(board, team, enemy,
                                                 search_depth, alpha, beta, False)[0]

                if eval_position > max_eval or best_move is None:
                    best_move = (piece, move)
                max_eval = max(max_eval, eval_position)
                alpha = max(alpha, eval_position)
                if eval_position >= beta:
                    break
                if beta <= alpha:
                    break
            return max_eval, best_move

        min_evaluation = math.inf
        min_move = None
        for piece, move in enemy.compute_valid_moves(board, team.pieces):
            search_depth = depth - 1
            search_depth = piece.increase_search_depth(search_depth)

            hash_ = compute_hash(board, self.hash_values)
            eval_position = self.board_hashes.get((maximizing_player, hash_))
            if eval_position is None:
                with ReversibleMove(board, piece, move.position, team.pieces):
                    eval_position = self.run(board, team, enemy,
                                             search_depth, alpha, beta, True)[0]

            if eval_position < min_evaluation or min_move is None:
                min_move = (piece, move.position)
            min_evaluation = min(min_evaluation, eval_position)
            beta = min(beta, eval_position)
            if eval_position <= alpha:
                break
            if beta <= alpha:
                break

        return min_evaluation, min_move



# -*- coding: utf-8 -*-
"""
Created on Mon Feb  8 12:56:04 2021

@author: Korean_Crimson
"""
import itertools
from dataclasses import dataclass
from typing import List
from typing import Tuple

@dataclass
class Move:
    """Moves class, takes a position that can be moved to and whether it is a capture or not"""

    position: Tuple[int, int]
    can_capture: bool = False

    def __hash__(self):
        return hash(self.position)

    def __iter__(self):
        #pylint: disable=invalid-name
        for x in self.position:
            yield x

    def __getitem__(self, key):
        return self.position[key]

# pylint: disable=invalid-name
@dataclass
class LineMove:
    """Move class. Encapsulates how a piece moves. To be inherited from when implementing a move"""

    range_: int
    direction: int = 1
    forward_only: bool = False
    horz_move: bool = False
    can_capture: bool = False

    def compute_valid_moves(self, board, piece) -> List[Move]:
        """Computes all valid moves that can be made from the passed position"""
        x, y = piece.position
        horz_moves = self._compute_horz_moves(x, y, board, piece.team)
        vert_moves = self._compute_vert_moves(x, y, board, piece.team)
        return horz_moves + vert_moves

    def _compute_horz_moves(self, current_x, y, board, team):
        if not self.horz_move:
            return []

        moves = []

        #forward
        for x in range(current_x + 1, board.size):
            square = board[x, y]
            if square is None:
                moves.append(Move((x, y)))
            elif square.team != team:
                moves.append(Move((x, y), can_capture=True))
            if square is not None:
                break

        #backward
        for x in range(current_x - 1, -1, -1):
            square = board[x, y]
            if square is None:
                moves.append(Move((x, y)))
            else:
                if square.team != team and self.can_capture:
                    moves.append(Move((x, y), can_capture=True))
                break

        return moves

    def _compute_vert_moves(self, x, current_y, board, team):
        forward_moves = []
        for y in range(current_y - 1, current_y - 1 - self.range_, -1):
            if y < 0:
                break

            square = board[x, y]
            if square is None:
                forward_moves.append(Move((x, y)))
            else:
                if square.team != team and self.can_capture:
                    forward_moves.append(Move((x, y), can_capture=True))
                break

        backward_moves = []
        for y in range(current_y + 1, current_y + 1 + self.range_):
            if y >= board.size:
                break

            square = board[x, y]
            if square is None:
                backward_moves.append(Move((x, y)))
            else:
                if square.team != team and self.can_capture:
                    backward_moves.append(Move((x, y), can_capture=True))
                break

        if self.forward_only:
            return forward_moves if self.direction == -1 else backward_moves
        return forward_moves + backward_moves

# pylint: disable=too-few-public-methods
class BishopMove:
    """Moves diagonally to either side of the board, backwards and forwards"""

    @staticmethod
    def compute_valid_moves(board, piece) -> List[Move]:
        """Computes all valid moves that can be made from the passed position"""
        moves = []
        x, y = piece.position
        team = piece.team
        for x_dir, y_dir in itertools.product((1, -1), repeat=2):
            pos = (x + x_dir, y + y_dir)
            while board.is_on_board(pos):
                square = board[pos]
                if square is None:
                    moves.append(Move(pos))
                elif square.team != team:
                    moves.append(Move(pos, can_capture=True))
                if square is not None:
                    break
                pos = (pos[0] + x_dir, pos[1] + y_dir)
        return moves

class InitialPawnMove:
    """Moves 2 spaces forward"""

    def __init__(self, direction):
        self.direction = direction

    def compute_valid_moves(self, board, piece) -> List[Move]:
        """Computes all valid moves that can be made from the passed position"""
        if len(piece.position_history):
            return []
        x, y = piece.position
        dir_ = self.direction
        def empty(y_):
            pos = (x, y + y_ * dir_)
            return board.is_empty_at(pos) and board.is_on_board(pos)
        return [Move((x, y + 2 * dir_))] if empty(1) and empty(2) else []

class KingMove:
    """Moves 1 space in any direction"""

    @staticmethod
    def compute_valid_moves(board, piece) -> List[Move]:
        """Computes all valid moves that can be made from the passed position"""
        neighbours = lambda x: [y for y in (x - 1, x, x + 1) if 0 <= y < board.size]
        x, y = piece.position
        moves = []
        for pos in itertools.product(neighbours(x), neighbours(y)):
            if pos == piece.position:
                continue
            square = board[pos]
            if square is None:
                moves.append(Move(pos))
            elif square.team != piece.team:
                moves.append(Move(pos, can_capture=True))
        return moves

class PawnMove:
    """Moves 1 space forward"""

    def __init__(self, direction):
        self.direction = direction

    def compute_valid_moves(self, board, piece) -> List[Move]:
        """Computes all valid moves that can be made from the passed position"""
        x, y = piece.position
        pos = (x, y + 1 * self.direction)
        if board.is_on_board(pos) and board.is_empty_at(pos):
            return [Move(pos)]
        return []

class PawnCapture:
    """Moves 1 space forward diagonally and needs to capture"""

    def __init__(self, direction):
        self.direction = direction

    def compute_valid_moves(self, board, piece) -> List[Move]:
        """Computes all valid moves that can be made from the passed position"""
        x, y = piece.position
        moves = ((x + 1, y + self.direction), (x - 1, y + self.direction))
        return [Move(x, can_capture=True) for x in filter(board.is_on_board, moves)
                if not board.is_empty_at(x) and board.is_enemy(x, piece.team)]

class EnPassantMove:
    """Pawn capture but with a pawn that already passed"""


class RookMove(LineMove):
    """Moves horizontally or vertically for up to 8 spaces"""

    def __init__(self):
        range_ = 8
        super().__init__(range_, horz_move=True, can_capture=True)


class KnightMove:
    """Knight move"""

    INDICES = [(x, y) for x, y in itertools.product((1, 2, -1, -2), repeat=2)
                   if abs(x) != abs(y)]

    @classmethod
    def compute_valid_moves(cls, board, piece) -> List[Move]:
        """Computes all valid moves that can be made from the passed position"""
        x1, y1 = piece.position
        positions = ((x1 + x2, y1 + y2) for x2, y2 in cls.INDICES)
        moves = []
        for pos in filter(board.is_on_board, positions):
            if board.is_empty_at(pos):
                moves.append(Move(pos))
            elif board.is_enemy(pos, piece.team):
                moves.append(Move(pos, can_capture=True))
        return moves



# -*- coding: utf-8 -*-
"""
Created on Mon Feb  8 12:43:27 2021

@author: Korean_Crimson
"""

PAWN = 'o'
KNIGHT = 'N'
BISHOP = 'B'
ROOK = 'R'
QUEEN = 'Q'
KING = 'K'

WHITE = '1'
BLACK = '2'

BACKROW = [ROOK, KNIGHT, BISHOP, QUEEN, KING, BISHOP, KNIGHT, ROOK]
FRONTROW = [PAWN] * 8

BOARD = [[piece + BLACK for piece in BACKROW],
         [piece + BLACK for piece in FRONTROW],
         [None] * 8,
         [None] * 8,
         [None] * 8,
         [None] * 8,
         [piece + WHITE for piece in FRONTROW],
         [piece + WHITE for piece in BACKROW]
         ]

DIRECTIONS = {WHITE: -1, BLACK: 1}
EARLY_VALUES = {PAWN: 4, KNIGHT: 5, BISHOP: 4, ROOK: 1, QUEEN: 3, KING: 2}
MID_VALUES = {PAWN: 2, KNIGHT: 3, BISHOP: 3, ROOK: 4, QUEEN: 6, KING: 10}
LATE_VALUES = {PAWN: 1, KNIGHT: 3, BISHOP: 4, ROOK: 5, QUEEN: 9, KING: 10}



# -*- coding: utf-8 -*-
"""
Created on Mon Feb  8 20:35:48 2021

@author: Korean_Crimson
"""
from typing import Tuple

X_POSITIONS = dict(zip(range(8), 'abcdefgh'))

#pylint: disable=invalid-name
def convert(position: Tuple[int, int]) -> str:
    """Converts an int index position on the board into a standard chess square name.
    Example: (0, 7) -> a1
    """
    x, y = position
    return X_POSITIONS[x] + str(8 - y)

def convert_str(string: str) -> Tuple[int, int]:
    """Converts a standard chess square name into an int index position on the board.
    Example: a1 -> (0, 7).
    """
    y = 8 - int(string[-1])
    for k, v in X_POSITIONS.items():
        if v == string[0]:
            x = k
            break
    return (x, y)

def is_diagonal(position1: Tuple[int, int], position2: Tuple[int, int]):
    """Returns True if position1 and position2 are on a diagonal line"""
    x1, y1 = position1
    x2, y2 = position2
    return abs(x1 - x2) == abs(y1 - y2)

def is_straight(position1: Tuple[int, int], position2: Tuple[int, int]):
    """Returns True if position1 and position2 are horizontally or vertically aligned"""
    x1, y1 = position1
    x2, y2 = position2
    return x1 == x2 or y1 == y2



# -*- coding: utf-8 -*-
"""
Created on Mon Feb  8 15:19:32 2021

@author: Korean_Crimson
"""
import itertools
import re
from typing import List
from typing import Tuple
from typing import Union

from colorama import Back
from colorama import Fore

from .consts import WHITE
from .piece import Piece

class Board:
    """Board class. Contains all the pieces on the chess board"""

    def __init__(self, pieces: List[Piece] = None, size = 8):
        self._pieces = {piece.position: piece for piece in pieces} if pieces is not None else {}
        self._squares = {pos: self._pieces.get(pos) for pos in
                         itertools.product(range(size), repeat=2)}
        self.size = size
        self.move_history: List[Tuple[Piece, Tuple[int, int]]] = []

    def __repr__(self):
        #pylint: disable=invalid-name
        for y in range(self.size):
            for x in range(self.size):
                if y % 2:
                    background = Back.BLACK if x % 2 else Back.WHITE
                else:
                    background = Back.WHITE if x % 2 else Back.BLACK

                piece = self[x, y]
                if piece is not None:
                    foreground = Fore.GREEN if WHITE in piece.representation else Fore.RED
                    representation = piece.representation
                else:
                    foreground = Fore.LIGHTBLACK_EX
                    representation = '  '

                square = re.sub('[12]', ' ', representation)
                print(background + foreground + square, end='')
            print()
        return ''

    def __iter__(self):
        #pylint: disable=invalid-name
        for y in range(self.size):
            yield [self[x, y] for x in range(self.size)]

    def __getitem__(self, value):
        return self._squares.get(value)

    def __setitem__(self, key, value):
        self._pieces[key] = value
        self._squares[key] = value

    def move_piece_and_capture(self, position, piece, enemy_pieces, log=True):
        """Moves piece to position on the board. Captures enemy if there is one"""
        captured_piece = None
        if not self.is_empty_at(position):
            if self[position] in enemy_pieces:
                captured_piece = self.capture_at(position, log=log)
                enemy_pieces.remove(captured_piece)
        self.move_piece(piece, position, log=log)
        return captured_piece

    def move_piece(self, piece: Piece, position: Tuple[int, int], log=True) -> None:
        """Moves the passed piece from the current position to the passed position"""
        self._pieces.pop(piece.position)
        self._squares[piece.position] = None
        piece.move_to(position, log=log)
        self._pieces[piece.position] = piece
        self._squares[piece.position] = piece
        self.move_history.append((piece, position))

    def capture_at(self, position: Tuple[int, int], log=True) -> Union[None, Piece]:
        """Removes the piece at the passed position and marks it as captured"""
        if not self.is_empty_at(position):
            piece = self._pieces.pop(position)
            self._squares[position] = None
            piece.captured = True
            if log:
                print(f'Captured {piece}')
            return piece
        return None

    def is_empty_at(self, position: Tuple[int, int]) -> bool:
        """Returns True if the checked position is None (contains no piece) else False"""
        return self[position] is None

    def is_on_board(self, position: Tuple[int, int]) -> bool:
        """Returns True if the checked position is on the board"""
        return position in self._squares

    def is_enemy(self, position: Tuple[int, int], team: int) -> bool:
        """Returns True if the checked position contains a piece with a team
        that is not the one specified.
        """
        return self[position] is not None and self[position].team != team

    def is_draw_by_repetition(self, repetitions=3, number_of_teams=2) -> bool:
        """Returns True if the teams have repeated the same moves a specified amount of times"""
        number_of_moves = repetitions * number_of_teams * 2
        if len(self.move_history) < number_of_moves:
            return False
        return len(set(self.move_history[-number_of_moves:])) == number_of_teams * 2



# -*- coding: utf-8 -*-
"""
Created on Mon Feb  8 13:33:57 2021

@author: Korean_Crimson
"""
import logging
import math
from typing import List

from chess_engine import move
from chess_engine.consts import BISHOP
from chess_engine.consts import KING
from chess_engine.consts import KNIGHT
from chess_engine.consts import PAWN
from chess_engine.consts import QUEEN
from chess_engine.consts import ROOK
from chess_engine.util import convert
from chess_engine.util import convert_str
from chess_engine.util import is_diagonal
from chess_engine.util import is_straight

from .move import Move

TURN_COUNTER = 0.0

#pylint: disable=invalid-name
class Piece:
    """Piece class. Needs to be subclassed by the various chess pieces"""

    def __init__(self, moves, position, representation):
        self.moves = moves
        self.position = convert_str(position) if isinstance(position, str) else position
        self.representation = representation
        self.team = representation[-1]
        self.captured = False
        self.position_history = []

    def __repr__(self):
        return f'{self.__class__.__name__}({convert(self.position)})'

    def __hash__(self):
        return hash((self.position, self.representation))

    def update(self, board): #pylint: disable=no-self-use
        """Update piece"""

    def increase_search_depth(self, search_depth): #pylint: disable=no-self-use
        """Does nothing to search depth"""
        return search_depth

    def compute_valid_moves(self, board) -> List[Move]:
        """Returns a list of squares that the piece can move to or capture at"""
        return [pos for move in self.moves for pos in move.compute_valid_moves(board, self)]

    def move_to(self, position, log=True) -> None:
        """Moves the piece to the specified position and adds it to the position history"""
        #pylint: disable=logging-fstring-interpolation
        pos_old = convert(self.position)
        self.position = position
        self.position_history.append(position)
        if log:
            global TURN_COUNTER #pylint: disable=global-statement
            TURN_COUNTER += 0.5
            logger = logging.getLogger('game.log')
            move_ = f'{self.representation} from {pos_old} to {convert(position)}'
            logger.info(f'Turn {math.ceil(TURN_COUNTER)}: Team {self.team}: {move_}')

    def can_capture_at(self, board, position) -> bool:
        """Returns true if this piece can move to the specified position.
        Note: doesnt check if piece is from opposite team.
        """
        return position in (move.position for move in self.compute_valid_moves(board))

class Pawn(Piece):
    """Pawn class. Contains all the pawn moves"""

    def __init__(self, direction, position, representation):
        self._unpromoted_moves = [move.InitialPawnMove(direction),
                 move.PawnMove(direction),
                 move.PawnCapture(direction)]
        self._promoted_moves = [move.RookMove(), move.BishopMove()]
        self.promoted = False
        self.direction = direction
        super().__init__(self._unpromoted_moves, position, representation)

    def update(self, board):
        self._update_promotion(board)

    def compute_valid_moves(self, board) -> List[Move]:
        """Returns a list of squares that the piece can move to or capture at"""
        self._update_promotion(board)
        return super().compute_valid_moves(board)

    def can_capture_at(self, board, position) -> bool:
        """Returns true if this piece can move to the specified position"""
        #optimization: return False if more than one square away
        if not self.promoted and abs(position[1] - self.position[1]) > 1:
            return False
        return super().can_capture_at(board, position)

    def _update_promotion(self, board):
        target_y = 0 if self.direction == -1 else board.size - 1
        if target_y in (y for _, y in self.position_history + [self.position]):
            self.promoted = True
            self.moves = self._promoted_moves
        else:
            self.promoted = False
            self.moves = self._unpromoted_moves

    @property
    def representation(self):
        """Representation getter, depends on the promoted state"""
        return f'{QUEEN}{self.team}' if self.promoted else f'{PAWN}{self.team}'

    @representation.setter
    def representation(self, value):
        pass

class Knight(Piece):
    """Knight class. Contains all the knight moves"""

    def __init__(self, _, position, representation):
        moves = [move.KnightMove()]
        super().__init__(moves, position, representation)

    def can_capture_at(self, board, position) -> bool:
        """Returns true if this piece can move to the specified position"""
        x1, y1 = position
        x2, y2 = self.position
        #optimization: return False if not L-shaped 2 and 1 squares away
        if {abs(x1 - x2), abs(y1 - y2)} != {1, 2}:
            return False
        return super().can_capture_at(board, position)

class Bishop(Piece):
    """Bishop class. Contains all the bishop moves"""

    def __init__(self, _, position, representation):
        moves = [move.BishopMove()]
        super().__init__(moves, position, representation)

    def can_capture_at(self, board, position) -> bool:
        """Returns true if this piece can move to the specified position"""
        #optimization: return False if not on same diagonal
        if not is_diagonal(position, self.position):
            return False
        return super().can_capture_at(board, position)

class Rook(Piece):
    """Rook class. Contains all the rook moves"""

    def __init__(self, _, position, representation):
        moves = [move.RookMove()]
        super().__init__(moves, position, representation)

    def can_capture_at(self, board, position) -> bool:
        """Returns true if this piece can move to the specified position"""
        #optimization: return False if not horizontally or vertically aligned
        if not is_straight(position, self.position):
            return False
        return super().can_capture_at(board, position)

class Queen(Piece):
    """Queen class. Contains all the queen moves"""

    def __init__(self, _, position, representation):
        moves = [move.RookMove(), move.BishopMove()]
        self.depth_counter = 3
        super().__init__(moves, position, representation)

    def increase_search_depth(self, search_depth):
        if self.depth_counter <= 0:
            return search_depth
        self.depth_counter -= 1
        return search_depth + 2

    def can_capture_at(self, board, position) -> bool:
        """Returns true if this piece can move to the specified position"""
        #optimization: return False if not horizontally or vertically aligned
        #and not same diagonal
        if (not is_straight(position, self.position)
            and not is_diagonal(position, self.position)):
            return False
        return super().can_capture_at(board, position)

class King(Piece):
    """King class. Contains all the king moves"""

    def __init__(self, _, position, representation):
        moves = [move.KingMove()]
        super().__init__(moves, position, representation)

    def can_capture_at(self, board, position) -> bool:
        """Returns true if this piece can move to the specified position"""
        x1, y1 = position
        x2, y2 = self.position
        #optimization: return False if more than one square away
        if abs(x1 - x2) > 1 or abs(y1 - y2) > 1:
            return False
        return super().can_capture_at(board, position)

PIECES = {PAWN: Pawn,
          KNIGHT: Knight,
          BISHOP: Bishop,
          ROOK: Rook,
          QUEEN: Queen,
          KING: King}



# -*- coding: utf-8 -*-
"""
Created on Fri Jan  7 14:54:55 2022

@author: richa
"""

from dataclasses import dataclass
from typing import Dict
from typing import List

from .algorithm import ReversibleMove
from .consts import EARLY_VALUES
from .piece import King
from .piece import Piece
from .piece import Queen

@dataclass
class Team:
    """Contains all the pieces for a side"""

    pieces: List[Piece]
    representation: str

    def __post_init__(self):
        self.king = [x for x in self.pieces if isinstance(x, King)][0]
        self.queen = [x for x in self.pieces if isinstance(x, Queen)][0]
        self.sort_pieces(EARLY_VALUES)

    def __repr__(self):
        return self.representation

    def sort_pieces(self, values: Dict[str, int]):
        """Sorts the pieces by the specified values dict"""
        self.pieces.sort(key=lambda x: values.get(x.representation[0]), reverse=True)

    def compute_all_moves(self, board):
        """Returns all valid moves for all pieces passed in"""
        return [(piece, move) for piece in self.pieces
                for move in piece.compute_valid_moves(board)]

    def compute_valid_moves(self, board, enemy_pieces):
        """Returns moves not resulting in a check of the allied king"""
        valid_moves = []
        for piece, move in self.compute_all_moves(board):
            with ReversibleMove(board, piece, move.position, enemy_pieces):
                if not self.in_check(board, enemy_pieces):
                    valid_moves.append((piece, move))
        #random.shuffle(valid_moves) #HACK to avoid the same game consistently
        valid_moves.sort(key=lambda x: x[1].can_capture, reverse=True)
        return valid_moves

    def in_check(self, board, enemy_pieces) -> bool:
        """Returns True if any enemy pieces can capture at the specified position"""
        #optimization: True for all truthy elements.
        #Directly checking capture state would be slower
        return any(True for x in enemy_pieces if x.can_capture_at(board, self.king.position))



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
from pathlib import Path

import setuptools

project_dir = Path(__file__).parent

setuptools.setup(
    name="chess_ng",
    version="1.0.0",
    description="Chess engine including minimax AI",
    long_description=project_dir.joinpath("README.md").read_text(encoding="utf-8"),
    long_description_content_type="text/markdown",
    keywords=["python"],
    author="Richard Baltrusch",
    url="https://github.com/rbaltrusch/chess_engine",
    packages=setuptools.find_packages("."),
    package_dir={"": "."},
    python_requires=">=3.8",
    include_package_data=True,
    package_data={"chess_engine": ["py.typed"]},  # for mypy
    # This is a trick to avoid duplicating dependencies between both setup.py and requirements.txt.
    # requirements.txt must be included in MANIFEST.in for this to work.
    install_requires=project_dir.joinpath("requirements.txt").read_text().split("\n"),
    zip_safe=False,
    license="MIT",
    classifiers=[
        "Development Status :: 4 - Beta",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3 :: Only",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Typing :: Typed",
    ],
)



# -*- coding: utf-8 -*-
"""
Created on Sat Jan 22 14:53:45 2022

@author: richa
"""
import math
from typing import Dict

def get_hash_values(pieces) -> Dict[str, int]:
    """Returns a hash value dict that looks up piece representations and returns
    a unique integer.
    """
    representations = set(piece.representation for piece in pieces)
    return {representation: i for i, representation in enumerate(representations, 1)}

def compute_hash(board, hash_values) -> int:
    """Computes a new hash from the board the specified hash_values lookup table"""
    #pylint: disable=invalid-name
    if not hash_values:
        return 0

    exponent = math.ceil(math.log(len(hash_values), 2))
    hash_ = 0
    for y in range(board.size):
        for x in range(board.size):
            counter = y * board.size + x
            piece = board[x, y]
            if piece is None:
                continue

            value = hash_values[piece.representation]
            num = value << counter * exponent
            hash_ ^= num
    return hash_

def update_hash(hash_, board, hash_values, changed_positions) -> int:
    """Updates the Zobrist hash by xor gating. 100 times fasteer than generating new hash"""
    #pylint: disable=invalid-name
    exponent = math.ceil(math.log(len(hash_values), 2))
    max_ = 2 ** exponent - 1
    for x, y in changed_positions:
        counter = y * board.size + x
        piece = board[x, y]
        value = hash_values[piece.representation] if piece is not None else 0
        num = value << counter * exponent
        hash_ &= ~(max_ << counter * exponent)
        hash_ ^= num
    return hash_
